import sys
from protocol import IcingProtocol

RED = "\033[91m"
GREEN = "\033[92m"
YELLOW = "\033[93m"
BLUE = "\033[94m"
RESET = "\033[0m"


def main():
    protocol = IcingProtocol()

    print(f"{YELLOW}\n======================================")
    print("   Icing Protocol - Secure Communication   ")
    print("======================================\n" + RESET)
    print(f"Listening on port: {protocol.local_port}")
    print(f"Your identity public key (hex): {protocol.identity_pubkey.hex()}")
    print("\nAvailable commands:")
    print("  help                       - Show this help message")
    print("  peer_id <hex_pubkey>       - Set peer identity public key")
    print("  connect <port>             - Connect to a peer at the specified port")
    print("  generate_ephemeral_keys    - Generate ephemeral ECDH keys")
    print("  send_ping [cipher]         - Send PING request (cipher: 0=AES-GCM, 1=ChaCha20-Poly1305, default: 0)")
    print("  respond_ping <index> <0|1> - Respond to a PING (0=reject, 1=accept)")
    print("  send_handshake             - Send handshake with ephemeral keys")
    print("  generate_ecdhe <index>     - Process handshake at specified index")
    print("  derive_hkdf                - Derive encryption key using HKDF")
    print("  send_encrypted <plaintext> - Encrypt and send a message")
    print("  decrypt <index>            - Decrypt received message at index")
    print("  auto_responder <on|off>    - Enable/disable automatic responses")
    print("  show_state                 - Display current protocol state")
    print("  exit                       - Exit the program\n")

    while True:
        try:
            line = input("Cmd> ").strip()
        except EOFError:
            break
        if not line:
            continue
            
        parts = line.split()
        cmd = parts[0].lower()
        
        try:
            if cmd == "exit":
                protocol.stop()
                break
                
            elif cmd == "help":
                print("\nAvailable commands:")
                print("  help                       - Show this help message")
                print("  peer_id <hex_pubkey>       - Set peer identity public key")
                print("  connect <port>             - Connect to a peer at the specified port")
                print("  generate_ephemeral_keys    - Generate ephemeral ECDH keys")
                print("  send_ping [cipher]         - Send PING request (cipher: 0=AES-GCM, 1=ChaCha20-Poly1305, default: 0)")
                print("  respond_ping <index> <0|1> - Respond to a PING (0=reject, 1=accept)")
                print("  send_handshake             - Send handshake with ephemeral keys")
                print("  generate_ecdhe <index>     - Process handshake at specified index")
                print("  derive_hkdf                - Derive encryption key using HKDF")
                print("  send_encrypted <plaintext> - Encrypt and send a message")
                print("  decrypt <index>            - Decrypt received message at index")
                print("  auto_responder <on|off>    - Enable/disable automatic responses")
                print("  show_state                 - Display current protocol state")
                print("  exit                       - Exit the program")
                
            elif cmd == "show_state":
                protocol.show_state()
                
            elif cmd == "peer_id":
                if len(parts) != 2:
                    print(f"{RED}[ERROR]{RESET} Usage: peer_id <hex_pubkey>")
                    continue
                try:
                    protocol.set_peer_identity(parts[1])
                except ValueError as e:
                    print(f"{RED}[ERROR]{RESET} Invalid public key: {e}")
                    
            elif cmd == "connect":
                if len(parts) != 2:
                    print(f"{RED}[ERROR]{RESET} Usage: connect <port>")
                    continue
                try:
                    port = int(parts[1])
                    protocol.connect_to_peer(port)
                except ValueError:
                    print(f"{RED}[ERROR]{RESET} Invalid port number.")
                except Exception as e:
                    print(f"{RED}[ERROR]{RESET} Connection failed: {e}")
                    
            elif cmd == "generate_ephemeral_keys":
                protocol.generate_ephemeral_keys()
                
            elif cmd == "send_ping":
                # Optional cipher parameter (0 = AES-GCM, 1 = ChaCha20-Poly1305)
                cipher = 0  # Default to AES-GCM
                if len(parts) >= 2:
                    try:
                        cipher = int(parts[1])
                        if cipher not in (0, 1):
                            print(f"{YELLOW}[WARNING]{RESET} Unsupported cipher code {cipher}. Using AES-GCM (0).")
                            cipher = 0
                    except ValueError:
                        print(f"{YELLOW}[WARNING]{RESET} Invalid cipher code. Using AES-GCM (0).")
                protocol.send_ping_request()
                
            elif cmd == "send_handshake":
                protocol.send_handshake()
                
            elif cmd == "respond_ping":
                if len(parts) != 3:
                    print(f"{RED}[ERROR]{RESET} Usage: respond_ping <index> <0|1>")
                    continue
                try:
                    idx = int(parts[1])
                    answer = int(parts[2])
                    if answer not in (0, 1):
                        print(f"{RED}[ERROR]{RESET} Answer must be 0 (reject) or 1 (accept).")
                        continue
                    protocol.respond_to_ping(idx, answer)
                except ValueError:
                    print(f"{RED}[ERROR]{RESET} Index and answer must be integers.")
                except Exception as e:
                    print(f"{RED}[ERROR]{RESET} Failed to respond to ping: {e}")
                    
            elif cmd == "generate_ecdhe":
                if len(parts) != 2:
                    print(f"{RED}[ERROR]{RESET} Usage: generate_ecdhe <index>")
                    continue
                try:
                    idx = int(parts[1])
                    protocol.generate_ecdhe(idx)
                except ValueError:
                    print(f"{RED}[ERROR]{RESET} Index must be an integer.")
                except Exception as e:
                    print(f"{RED}[ERROR]{RESET} Failed to process handshake: {e}")
                    
            elif cmd == "derive_hkdf":
                try:
                    protocol.derive_hkdf()
                except Exception as e:
                    print(f"{RED}[ERROR]{RESET} Failed to derive HKDF key: {e}")
                    
            elif cmd == "send_encrypted":
                if len(parts) < 2:
                    print(f"{RED}[ERROR]{RESET} Usage: send_encrypted <plaintext>")
                    continue
                plaintext = " ".join(parts[1:])
                try:
                    protocol.send_encrypted_message(plaintext)
                except Exception as e:
                    print(f"{RED}[ERROR]{RESET} Failed to send encrypted message: {e}")
                    
            elif cmd == "decrypt":
                if len(parts) != 2:
                    print(f"{RED}[ERROR]{RESET} Usage: decrypt <index>")
                    continue
                try:
                    idx = int(parts[1])
                    protocol.decrypt_received_message(idx)
                except ValueError:
                    print(f"{RED}[ERROR]{RESET} Index must be an integer.")
                except Exception as e:
                    print(f"{RED}[ERROR]{RESET} Failed to decrypt message: {e}")
                    
            elif cmd == "auto_responder":
                if len(parts) != 2:
                    print(f"{RED}[ERROR]{RESET} Usage: auto_responder <on|off>")
                    continue
                val = parts[1].lower()
                if val not in ("on", "off"):
                    print(f"{RED}[ERROR]{RESET} Value must be 'on' or 'off'.")
                    continue
                protocol.enable_auto_responder(val == "on")
                
            else:
                print(f"{RED}[ERROR]{RESET} Unknown command: {cmd}")
                print("Type 'help' for a list of available commands.")
                
        except Exception as e:
            print(f"{RED}[ERROR]{RESET} Command failed: {e}")

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("\nExiting...")
    except Exception as e:
        print(f"{RED}[FATAL ERROR]{RESET} {e}")
        sys.exit(1)
